/*
 *        spAudio.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spFile.h>
#include <sp/spOption.h>

#include <sp/spPluginP.h>

#include <sp/spAudioP.h>
#include "spAudio_Plugin.h"

typedef struct _spAudioDeviceList {
    char *driver_name;

    int device_index;
    char *device_name;
    
    struct _spAudioDeviceList *prev;
    struct _spAudioDeviceList *next;
} *spAudioDeviceList;    


void spTerminateAudio(spAudio audio);

int spGetNumAudioDriverFromHostData(spPluginHostData host_data)
{
    int num_arch_driver = 0;
    int num_plugin_driver = 0;
    
    num_arch_driver = spGetNumAudioDriverArch();
    num_plugin_driver = spGetNumAudioDriverPluginFromHostData(host_data);
    
    spDebug(20, "spGetNumAudioDriver", "num_arch_driver = %d, num_plugin_driver = %d\n",
            num_arch_driver, num_plugin_driver);

    return num_arch_driver + num_plugin_driver;
}

int spGetNumAudioDriver(void)
{
    return spGetNumAudioDriverFromHostData(spGetPluginHostData());
}

char *xspGetAudioDriverNameFromHostData(spPluginHostData host_data, int index)
{
    int num_arch_driver;
    const char *name = NULL;

    if (index < 0) return NULL;
    
    num_arch_driver = spGetNumAudioDriverArch();
    spDebug(20, "xspGetAudioDriverName", "index = %d, num_arch_driver = %d\n",
            index, num_arch_driver);
    
    if (index < num_arch_driver) {
        name = spGetAudioDriverNameArch(index);
    } else {
        index -= num_arch_driver;
        name = spGetAudioDriverNamePluginFromHostData(host_data, index);
    }

    if (name != NULL) {
        spDebug(20, "xspGetAudioDriverName", "done: name = %s\n", name);
        return strclone(name);
    } else {
        return NULL;
    }
}

char *xspGetAudioDriverName(int index)
{
    return xspGetAudioDriverNameFromHostData(spGetPluginHostData(), index);
}

static spAudioDeviceList sp_audio_device_list = NULL;

static void destroyAudioDeviceList(void *data)
{
    spAudioDeviceList device_list, prev_device_list;
    
    device_list = sp_audio_device_list;
        
    while (device_list != NULL) {
        if (device_list->driver_name != NULL) {
            xfree(device_list->driver_name);
        }
        if (device_list->device_name != NULL) {
            xfree(device_list->device_name);
        }

        prev_device_list = device_list;
        device_list = device_list->next;

        xfree(prev_device_list);
    }

    sp_audio_device_list = NULL;

    return;
}

static spAudioDeviceList createAudioDeviceListFromHostData(spPluginHostData host_data)
{
    int k, l;
    int num_driver;
    int num_device;
    char *driver_name;
    spAudio audio;
    spAudioDeviceList device_list;
    spAudioDeviceList prev_device_list;

    if (sp_audio_device_list == NULL) {
        num_driver = spGetNumAudioDriverFromHostData(host_data);
        spDebug(20, "createAudioDeviceList", "num_driver = %d\n", num_driver);
        
        prev_device_list = NULL;
        for (k = 0; k < num_driver; k++) {
            driver_name = xspGetAudioDriverNameFromHostData(host_data, k);
            spDebug(20, "createAudioDeviceList", "k = %d, driver name = %s\n", k, driver_name);

            if ((audio = spInitAudioDriver(driver_name)) != NULL) {
                if (spGetNumAudioDevice(audio, &num_device) == SP_TRUE) {
                    spDebug(20, "createAudioDeviceList", "%s: num_device = %d\n", driver_name, num_device);
                    
                    for (l = 0; l < num_device; l++) {
                        device_list = xalloc(1, struct _spAudioDeviceList);
                        device_list->next = NULL;
                        device_list->prev = prev_device_list;
                        if (device_list->prev != NULL) {
                            device_list->prev->next = device_list;
                        }
                
                        device_list->driver_name = strclone(driver_name);
            
                        device_list->device_index = l;
                        device_list->device_name = xspGetAudioDeviceName(audio, l);
                        spDebug(20, "createAudioDeviceList", "%s No.%d: device name = %s\n",
                                driver_name, l, device_list->device_name);

                        if (sp_audio_device_list == NULL) {
                            sp_audio_device_list = device_list;
                        }

                        prev_device_list = device_list;
                    }
                }

                spFreeAudioDriver(audio);
            }

            xfree(driver_name);
        }

        spAddExitCallback((spExitCallbackFunc)destroyAudioDeviceList, NULL);
    }

    return sp_audio_device_list;
}

/* if driver_name == NULL, get all device */
int spGetNumAudioDriverDeviceFromHostData(spPluginHostData host_data, const char *driver_name)
{
    int num_device;
    spAudioDeviceList device_list;

    device_list = createAudioDeviceListFromHostData(host_data);

    num_device = 0;
    while (device_list != NULL) {
        if (strnone(driver_name)
            || streq(driver_name, device_list->driver_name)) {
            num_device++;
        }
        device_list = device_list->next;
    }

    return num_device;
}

int spGetNumAudioDriverDevice(const char *driver_name)
{
    return spGetNumAudioDriverDeviceFromHostData(spGetPluginHostData(), driver_name);
}

char *xspGetAudioDriverDeviceNameFromHostData(spPluginHostData host_data, const char *driver_name, int index)
{
    int num_device;
    spAudioDeviceList device_list;

    device_list = createAudioDeviceListFromHostData(host_data);

    num_device = 0;
    while (device_list != NULL) {
        if (strnone(driver_name)
            || streq(driver_name, device_list->driver_name)) {
            if (num_device == index) {
                if (strnone(driver_name)) {
                    char *buf;
                    int size;
                    size = (int)(strlen(device_list->driver_name) + strlen(device_list->device_name)) + 2;
                    buf = xalloc(size, char);
                    spStrCopy(buf, size, device_list->driver_name);
                    spStrCat(buf, size, "/");
                    spStrCat(buf, size, device_list->device_name);
                    return buf;
                } else {
                    return strclone(device_list->device_name);
                }
            }
            num_device++;
        }
        device_list = device_list->next;
    }

    return NULL;

}

char *xspGetAudioDriverDeviceName(const char *driver_name, int index)
{
    return xspGetAudioDriverDeviceNameFromHostData(spGetPluginHostData(), driver_name, index);
}

static spBool findAudioDriverFromHostData(spPluginHostData host_data, const char *name, spBool *arch_driver_flag, char *name_buf, int buf_size)
{
    int k;
    int index;
    int num_driver;
    int num_arch_driver;
    int num_plugin_driver;
    static char **driver_list = NULL;
    static int prev_num_arch_driver = -1;
    static int prev_num_plugin_driver = -1;

    spDebug(80, "findAudioDriver", "in\n");
    
    num_arch_driver = spGetNumAudioDriverArch();
    num_plugin_driver = spGetNumAudioDriverPluginFromHostData(host_data);
    num_driver = num_arch_driver + num_plugin_driver;
    spDebug(80, "findAudioDriver", "num_arch_driver = %d/%d, num_plugin_driver = %d/%d\n",
            num_arch_driver, prev_num_arch_driver, num_plugin_driver, prev_num_plugin_driver);

    if (num_arch_driver != prev_num_arch_driver
        || num_plugin_driver != prev_num_plugin_driver
        || driver_list == NULL) {
        if (driver_list != NULL) {
            for (k = 0; k < prev_num_arch_driver + prev_num_plugin_driver; k++) {
                xfree(driver_list[k]);
            }
            xfree(driver_list);
        }
        
        driver_list = xalloc(num_driver, char *);
        for (k = 0; k < num_driver; k++) {
            driver_list[k] = xspGetAudioDriverNameFromHostData(host_data, k);
        }

        prev_num_arch_driver = num_arch_driver;
        prev_num_plugin_driver = num_plugin_driver;
    }

#if defined(linux) && !defined(ANDROID)
    spDebug(80, "findAudioDriver", "linux, num_arch_driver = %d, num_plugin_driver = %d\n",
            num_arch_driver, num_plugin_driver);
    
    if (strnone(name) && num_arch_driver <= 0) {
        const char *p;
        name = NULL;
        for (k = 0; k < num_plugin_driver; k++) {
            p = spGetAudioDriverNamePluginFromHostData(host_data, k);
            spDebug(80, "findAudioDriver", "k = %d / %d, p = %s\n", k, num_plugin_driver, p);
            if (streq(p, "pulsesimple")) {
                name = p;
                *arch_driver_flag = SP_FALSE;
                spDebug(80, "findAudioDriver", "pulsesimple found\n");
                break;
            }
        }
    }
#endif
        
    if (strnone(name)) {
        spDebug(80, "findAudioDriver", "name == NULL, return 0\n");
        *arch_driver_flag = SP_TRUE;
        name_buf[0] = NUL;
        return SP_TRUE;
    }

    index = -1;
    for (k = 0; k < num_driver; k++) {
        if (streq(driver_list[k], name)) {
            if (k < num_arch_driver) {
                *arch_driver_flag = SP_TRUE;
                index = k;
            } else {
                *arch_driver_flag = SP_FALSE;
                index = k - num_arch_driver;
            }

            if (name != name_buf) {
                spStrCopy(name_buf, buf_size, name);
            }
        }
    }

    spDebug(80, "findAudioDriver", "index = %d, arch_driver_flag = %d\n",
            index, *arch_driver_flag);

    if (index >= 0) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

static void setAudioFunc(spAudio audio, spBool arch_driver_flag)
{
    if (arch_driver_flag == SP_TRUE) {
        audio->func_list.init_driver = spInitAudioDriverArch;
        audio->func_list.free_driver = spFreeAudioDriverArch;
        audio->func_list.terminate = spTerminateAudioArch;
    
        audio->func_list.get_num_device = spGetNumAudioDeviceArch;
        audio->func_list.get_device_name = spGetAudioDeviceNameArch;
        audio->func_list.select_device = spSelectAudioDeviceArch;
        
        audio->func_list.get_caps = NULL;
        audio->func_list.display_option_dialog = NULL;
        
        audio->func_list.set_buffer_size = spSetAudioBufferSizeArch;
        audio->func_list.set_num_buffer = spSetAudioNumBufferArch;
        audio->func_list.set_blocking_mode = spSetAudioBlockingModeArch;
        audio->func_list.set_callback_func = spSetAudioCallbackFuncArch;
    
        audio->func_list.set_sample_bit = spSetAudioSampleBitArch;
        audio->func_list.set_sample_rate = spSetAudioSampleRateArch;
        audio->func_list.set_channel = spSetAudioChannelArch;
        
        audio->func_list.open_device = spOpenAudioDeviceArch;
        audio->func_list.close_device = spCloseAudioDeviceArch;
        audio->func_list.get_output_position = spGetAudioOutputPositionArch;
        audio->func_list.stop = spStopAudioArch;
    
        audio->func_list.write = spWriteAudioArch;
        audio->func_list.sync = spSyncAudioArch;
    
        audio->func_list.read = spReadAudioArch;
    } else {
        audio->func_list.init_driver = spInitAudioDriverPlugin;
        audio->func_list.free_driver = spFreeAudioDriverPlugin;
        audio->func_list.terminate = spTerminateAudioPlugin;
    
        audio->func_list.get_num_device = spGetNumAudioDevicePlugin;
        audio->func_list.get_device_name = spGetAudioDeviceNamePlugin;
        audio->func_list.select_device = spSelectAudioDevicePlugin;
        
        audio->func_list.get_caps = spGetAudioCapsPlugin;
        audio->func_list.display_option_dialog = spDisplayAudioOptionDialogPlugin;
        
        audio->func_list.set_buffer_size = spSetAudioBufferSizePlugin;
        audio->func_list.set_num_buffer = spSetAudioNumBufferPlugin;
        audio->func_list.set_blocking_mode = spSetAudioBlockingModePlugin;
        audio->func_list.set_callback_func = spSetAudioCallbackFuncPlugin;
    
        audio->func_list.set_sample_bit = spSetAudioSampleBitPlugin;
        audio->func_list.set_sample_rate = spSetAudioSampleRatePlugin;
        audio->func_list.set_channel = spSetAudioChannelPlugin;
        
        audio->func_list.open_device = spOpenAudioDevicePlugin;
        audio->func_list.close_device = spCloseAudioDevicePlugin;
        audio->func_list.get_output_position = spGetAudioOutputPositionPlugin;
        audio->func_list.stop = spStopAudioPlugin;
    
        audio->func_list.write = spWriteAudioPlugin;
        audio->func_list.sync = spSyncAudioPlugin;
    
        audio->func_list.read = spReadAudioPlugin;
    }

    return;
}

spAudio spInitAudioDriverFromHostData(spPluginHostData host_data, const char *driver_name)
{
    spAudio audio;
    spBool arch_driver_flag = SP_TRUE;
    char driver_name_buf[SP_MAX_LINE];

    spDebug(50, "spInitAudioDriver", "in\n");
    
    audio = xalloc(1, struct _spAudio);
    memset(audio, 0, sizeof(struct _spAudio));
    audio->samp_bit = 16;
    audio->samp_rate = 8000;
    audio->num_channel = 1;
    audio->buffer_size = SP_DEFAULT_AUDIO_BUFFER_SIZE;
    audio->samp_byte = 2;
    audio->duplex_flag = 0;
    audio->only_flag = 0;
    audio->block_mode = SP_AUDIO_BLOCKING;
    audio->specified_samp_bit = 16;
    audio->input_total = 0;
    audio->output_total = 0;
    audio->num_buffer = SP_DEFAULT_NUM_BUFFER;

    audio->internal_buffer = NULL;
    audio->internal_buffer_size = 0;

    audio->call_type = 0;
    audio->supported_call_type = 0;
    audio->call_func = NULL;
    audio->call_data = NULL;
    audio->plugin = NULL;

    strcpy(audio->default_device_name, "");
    if (strnone(driver_name)) {
        strcpy(driver_name_buf, "");
    } else {
        char *p;
        
        strcpy(driver_name_buf, driver_name);

        if ((p = strchr(driver_name_buf, '/')) != NULL) {
            spStrCopy(audio->default_device_name, SP_MAX_LINE, p + 1);
            *p = NUL;
            spDebug(20, "spInitAudioDriver", "driver_name = %s, default_device_name = %s\n",
                    driver_name_buf, audio->default_device_name);
        }
    }
    
    if (findAudioDriverFromHostData(host_data, driver_name_buf, &arch_driver_flag,
                                    driver_name_buf, sizeof(driver_name_buf)) == SP_TRUE) {
        setAudioFunc(audio, arch_driver_flag);
        spDebug(50, "spInitAudioDriver", "driver_name_buf = %s, arch_driver_flag = %d\n",
                driver_name_buf, arch_driver_flag);

        if (audio->func_list.init_driver(audio, driver_name_buf, host_data) == SP_TRUE) {
            spAddExitCallback((spExitCallbackFunc)spTerminateAudio, audio);

            audio->internal_buffer_size = SP_AUDIO_DEFAULT_INTERNAL_BUFFER_SIZE;
            audio->internal_buffer = xalloc(audio->internal_buffer_size, char);
    
            spDebug(50, "spInitAudio", "done\n");
    
            return audio;
        }
    }

    spDebug(50, "spInitAudioDriver", "init driver failed\n");
    xfree(audio);
    return NULL;
}

spAudio spInitAudioDriver(const char *driver_name)
{
    spsetstdio(stdin, stdout, stderr);
    return spInitAudioDriverFromHostData(spGetPluginHostData(), driver_name);
}

spAudio spInitAudio(void)
{
    return spInitAudioDriver(NULL);
}

spBool spGetNumAudioDevice(spAudio audio, int *num_device)
{
    if (audio == NULL) return SP_FALSE;
    
    return audio->func_list.get_num_device(audio, num_device);
}

char *xspGetAudioDeviceName(spAudio audio, int device_index)
{
    const char *name;
    
    if (audio == NULL) return NULL;
    
    name = audio->func_list.get_device_name(audio, device_index);
    
    if (name != NULL) {
        return strclone(name);
    } else {
        return NULL;
    }
}

spBool spSelectAudioDevice(spAudio audio, int device_index)
{
    if (audio == NULL) return SP_FALSE;

    strcpy(audio->default_device_name, "");
    
    return audio->func_list.select_device(audio, device_index);
}

spBool spIsAudioOptionDialogExist(spAudio audio)
{
    unsigned long caps;

    if (audio == NULL || audio->func_list.get_caps == NULL) return SP_FALSE;

    caps = audio->func_list.get_caps(audio);
    
    if (caps & SP_AUDIO_CAPS_OPTION_DIALOG) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spDisplayAudioOptionDialog(spAudio audio)
{
    if (audio == NULL || audio->func_list.display_option_dialog == NULL) return SP_FALSE;

    spDebug(50, "spDisplayAudioOptionDialog", "in\n");
    
    return audio->func_list.display_option_dialog(audio);
}

spBool spSetAudioSampleRate(spAudio audio, double samp_rate)
{
    if (audio == NULL) return SP_FALSE;
    
    audio->samp_rate = samp_rate;
    spDebug(30, NULL, "sampling rate = %f\n", audio->samp_rate);
    
    return audio->func_list.set_sample_rate(audio);
}

spBool spGetAudioSampleRate(spAudio audio, double *samp_rate)
{
    if (audio == NULL) return SP_FALSE;

    if (samp_rate != NULL) {
        *samp_rate = audio->samp_rate;
    }
    
    return SP_TRUE;
}

spBool spSetAudioChannel(spAudio audio, int num_channel)
{
    if (audio == NULL || num_channel <= 0) return SP_FALSE;
    
    audio->num_channel = num_channel;
    spDebug(30, NULL, "channels = %d\n", audio->num_channel);
    
    return audio->func_list.set_channel(audio);
}

spBool spGetAudioChannel(spAudio audio, int *num_channel)
{
    if (audio == NULL) return SP_FALSE;

    if (num_channel != NULL) {
        *num_channel = audio->num_channel;
    }
    
    return SP_TRUE;
}

spBool spSetAudioBufferSize(spAudio audio, int buffer_size)
{
    if (audio == NULL) return SP_FALSE;

    if (buffer_size <= 0) {
        audio->buffer_size = SP_DEFAULT_AUDIO_BUFFER_SIZE;
    } else {
        audio->buffer_size = buffer_size;
    }
    spDebug(30, NULL, "audio buffer size = %d\n", audio->buffer_size);
    
    return audio->func_list.set_buffer_size(audio);
}

spBool spGetAudioBufferSize(spAudio audio, int *buffer_size)
{
    if (audio == NULL) return SP_FALSE;

    if (buffer_size != NULL) {
        *buffer_size = audio->buffer_size;
    }
    
    return SP_TRUE;
}

spBool spSetAudioNumBuffer(spAudio audio, int num_buffer)
{
    if (audio == NULL) return SP_FALSE;

    if (num_buffer <= 0) {
        audio->num_buffer = SP_DEFAULT_NUM_BUFFER;
    } else {
        audio->num_buffer = MAX(num_buffer, 2);
    }
    spDebug(30, NULL, "number of audio buffers = %d\n", audio->num_buffer);
    
    return audio->func_list.set_num_buffer(audio);
}

spBool spGetAudioNumBuffer(spAudio audio, int *num_buffer)
{
    if (audio == NULL) return SP_FALSE;

    if (num_buffer != NULL) {
        *num_buffer = audio->num_buffer;
    }
    
    return SP_TRUE;
}

spBool spSetAudioBlockingMode(spAudio audio, int block_mode)
{
    if (audio == NULL || block_mode < 0 || block_mode > 1) return SP_FALSE;
    
    audio->block_mode = block_mode;
    spDebug(30, NULL, "block_mode = %d\n", audio->block_mode);
    
    return audio->func_list.set_blocking_mode(audio);
}

spBool spGetAudioBlockingMode(spAudio audio, int *block_mode)
{
    if (audio == NULL) return SP_FALSE;

    if (block_mode != NULL) {
        *block_mode = audio->block_mode;
    }
    
    return SP_TRUE;
}

spBool spSetAudioSampleBit(spAudio audio, int samp_bit)
{
    if (audio == NULL || samp_bit < 8) return SP_FALSE;

    audio->specified_samp_bit = samp_bit;
    audio->samp_bit = samp_bit;
    return audio->func_list.set_sample_bit(audio);
}

spBool spGetAudioSampleBit(spAudio audio, int *samp_bit)
{
    if (audio == NULL) return SP_FALSE;

    if (samp_bit != NULL) {
        *samp_bit = audio->samp_bit;
    }
    
    return SP_TRUE;
}

spBool spGetAudioSpecifiedSampleBit(spAudio audio, int *specified_samp_bit)
{
    if (audio == NULL) return SP_FALSE;

    if (specified_samp_bit != NULL) {
        *specified_samp_bit = audio->specified_samp_bit;
    }
    
    return SP_TRUE;
}

spAudioCallbackType spSetAudioCallbackFunc(spAudio audio, spAudioCallbackType call_type,
                                           spAudioCallbackFunc call_func, void *call_data)
{
    if (audio == NULL) return 0;

    audio->call_func = call_func;
    audio->call_data = call_data;

    audio->call_type = call_type;
    audio->supported_call_type = audio->func_list.set_callback_func(audio);
    spDebug(30, "spSetAudioCallbackFunc", "call_type = %ld, supported_call_type = %ld\n",
            audio->call_type, audio->supported_call_type);

    return audio->call_type;
}

spBool spOpenAudioDeviceArch(spAudio audio, const char *mode)
{
    if (strlen(mode) >= 2 && mode[1] == 'o') {
        audio->only_flag = 1;
    }
    
    spDebug(50, "spOpneAudioDeviceArch", "mode = %s, only_flag = %d\n", mode, audio->only_flag);
    
    if (mode[0] == 'r') {
        audio->input_total = 0;
        return spOpenInputAudioDeviceArch(audio);
    } else if (mode[0] == 'w') {
        audio->output_total = 0;
        spDebug(50, "spOpneAudioDeviceArch", "call spOpenOutputAudioDeviceArch\n");
        return spOpenOutputAudioDeviceArch(audio);
    } else {
        spDebug(1, "spOpneAudioDeviceArch", "Unknown mode: %s\n", mode);
        return SP_FALSE;
    }
}

spBool spOpenAudioDevice(spAudio audio, const char *mode)
{
    if (audio == NULL || strnone(mode)) return SP_FALSE;

    spDebug(20, "spOpenAudioDevice", "mode = %s\n", mode);

    if (!strnone(audio->default_device_name)) {
        int i;
        int num_device;
        const char *name;
        
        spDebug(20, "spOpenAudioDevice", "audio->default_device_name = %s\n", audio->default_device_name);
        
        if (spGetNumAudioDevice(audio, &num_device) == SP_TRUE) {
            for (i = 0; i < num_device; i++) {
                name = audio->func_list.get_device_name(audio, i);
                if (name != NULL && streq(name, audio->default_device_name)) {
                    spSelectAudioDevice(audio, i);
                    break;
                }
            }
        }
    }

    return audio->func_list.open_device(audio, mode);
}

spBool spCloseAudioDevice(spAudio audio)
{
    if (audio == NULL) return SP_FALSE;
    
    return audio->func_list.close_device(audio);
}

static spBool isUnsupported32bit(spAudio audio)
{
    if ((audio->specified_samp_bit == 24 || audio->specified_samp_bit == 32)
        && audio->samp_bit != audio->specified_samp_bit) {
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool isUnsupportedFloat(spAudio audio)
{
    if ((audio->specified_samp_bit >= 33)
        && audio->samp_bit != audio->specified_samp_bit) {
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

static long readAudio16ToFloat(spAudio audio, char *data, long length)
{
    long k;
    long nread;
    short *sdata;
    float *fdata;
    double *ddata;
    double factor;
    
    factor = 32768.0;
    
    if ((nread = audio->func_list.read(audio, data, length)) > 0) {
        sdata = (short *)data;

        if (audio->specified_samp_bit >= 64) {
            ddata = (double *)data;
            
            for (k = length - 1; k >= 0; k--) {
                ddata[k] = (double)sdata[k] / factor;
            }
        } else {
            fdata = (float *)data;
            
            for (k = length - 1; k >= 0; k--) {
                fdata[k] = (float)((double)sdata[k] / factor);
            }
        }
    }

    return nread;
}

static long readAudio16To32(spAudio audio, char *data, long length)
{
    long k;
    long nread;
    short *sdata;
    long *ldata;
    long factor;
    
    if ((nread = audio->func_list.read(audio, data, length)) > 0) {
        sdata = (short *)data;
        ldata = (long *)data;

        if (audio->specified_samp_bit == 24) {
            factor = 256;
        } else {
            factor = 65536;
        }
        for (k = length - 1; k >= 0; k--) {
            ldata[k] = (long)sdata[k] * factor;
        }
    }

    return nread;
}

long _spReadAudio(spAudio audio, char *data, long length)
{
    long nread;
    
    if (audio == NULL || data == NULL || length < 0) return -1;

    if (isUnsupportedFloat(audio) == SP_TRUE) {
        nread = readAudio16ToFloat(audio, data, length);
    } else if (isUnsupported32bit(audio) == SP_TRUE) {
        nread = readAudio16To32(audio, data, length);
    } else {
        nread = audio->func_list.read(audio, data, length);
    }

    if (nread >= 0) {
        audio->input_total += (spULong)nread;
    }

    return nread;
}

long spReadAudio(spAudio audio, void *data, long length)
{
    return _spReadAudio(audio, (char *)data, length);
}

long spReadAudioBuffer(spAudio audio, char *buffer, long size)
{
    long samp_byte;
    long length;
    long nread;
    
    if (audio == NULL || buffer == NULL || size < 0) return -1;

    samp_byte = audio->specified_samp_bit / 8;
    length = size / samp_byte;

    nread = _spReadAudio(audio, buffer, length);

    return nread > 0 ? nread * samp_byte : nread;
}

static long writeAudioFloatTo16(spAudio audio, char *data, long length)
{
    long k;
    long offset;
    long nwrite;
    long internal_buffer_length;
    short *sdata;
    float *fdata;
    double *ddata;
    double factor;

    if (audio->internal_buffer == NULL) return -1;

    fdata = (float *)data;
    ddata = (double *)data;
    sdata = (short *)audio->internal_buffer;

    internal_buffer_length = audio->internal_buffer_size / 2;
    factor = 32768.0;
    
    offset = 0;
    while (offset < length) {
        nwrite = MIN(internal_buffer_length, length - offset);

        if (audio->specified_samp_bit >= 64) {
            for (k = 0; k < nwrite; k++) {
                sdata[k] = (short)spRound(ddata[offset + k] * factor);
            }
        } else {
            for (k = 0; k < nwrite; k++) {
                sdata[k] = (short)spRound((double)fdata[offset + k] * factor);
            }
        }
        
        if ((nwrite = audio->func_list.write(audio, (char *)sdata, nwrite)) <= 0) {
            break;
        }
        offset += nwrite;
    }

    return offset;
}

static long writeAudio32To16(spAudio audio, char *data, long length)
{
    long k;
    long offset;
    long nwrite;
    short *sdata;
    long *ldata;
    double factor;

    if (audio->internal_buffer == NULL) return -1;
    
    ldata = (long *)data;
    sdata = (short *)audio->internal_buffer;

    if (audio->specified_samp_bit == 24) {
        factor = 256.0;
    } else {
        factor = 65536.0;
    }
    
    offset = 0;
    while (offset < length) {
        nwrite = MIN(audio->internal_buffer_size / 2, length - offset);

        for (k = 0; k < nwrite; k++) {
            sdata[k] = (short)spRound((double)ldata[offset + k] / factor);
        }
        if ((nwrite = audio->func_list.write(audio, (char *)sdata, nwrite)) <= 0) {
            break;
        }
        offset += nwrite;
    }

    return offset;
}

long _spWriteAudio(spAudio audio, char *data, long length)
{
    long nwrite;
    
    if (audio == NULL || data == NULL || length < 0) return -1;

    if (isUnsupportedFloat(audio) == SP_TRUE) {
        nwrite = writeAudioFloatTo16(audio, data, length);
    } else if (isUnsupported32bit(audio) == SP_TRUE) {
        nwrite = writeAudio32To16(audio, data, length);
    } else {
        nwrite = audio->func_list.write(audio, data, length);
    }

    if (nwrite >= 0) {
        audio->output_total += (spULong)nwrite;
        
        if (audio->call_func != NULL
            && (audio->call_type & SP_AUDIO_OUTPUT_POSITION_CALLBACK)
            && !(audio->supported_call_type & SP_AUDIO_OUTPUT_POSITION_CALLBACK)) {
            spLong position;
            
            if (audio->func_list.get_output_position(audio, &position) == SP_TRUE) {
                audio->call_func(audio, SP_AUDIO_OUTPUT_POSITION_CALLBACK,
                                 &position, NULL, audio->call_data);
            }
        }
        if (audio->call_func != NULL
            && (audio->call_type & SP_AUDIO_OUTPUT_BUFFER_CALLBACK)
            && !(audio->supported_call_type & SP_AUDIO_OUTPUT_BUFFER_CALLBACK)) {
            audio->call_func(audio, SP_AUDIO_OUTPUT_BUFFER_CALLBACK,
                             data, &length, audio->call_data);
        }
    }

    return nwrite;
}

long spWriteAudio(spAudio audio, void *data, long length)
{
    return _spWriteAudio(audio, (char *)data, length);
}

long spWriteAudioBuffer(spAudio audio, char *buffer, long size)
{
    long samp_byte;
    long length;
    long nwrite;
    
    if (audio == NULL || buffer == NULL || size < 0) return -1;

    samp_byte = audio->specified_samp_bit / 8;
    length = size / samp_byte;

    nwrite = _spWriteAudio(audio, buffer, length);
    
    return nwrite > 0 ? nwrite * samp_byte : nwrite;
}

long spReadAudioDoubleWeighted(spAudio audio, double *data, long length, double weight)
{
    long k;
    long nread;
    short *sdata;
    long *ldata;
    float *fdata;
    double factor;

    spDebug(80, "spReadAudioDoubleWeighted", "length = %ld, weight = %f\n", length, weight);
    
    if ((nread = audio->func_list.read(audio, (char *)data, length)) > 0) {
        spDebug(80, "spReadAudioDoubleWeighted", "nread = %ld\n", nread);
        
        sdata = (short *)data;
        ldata = (long *)data;
        fdata = (float *)data;

        if (audio->samp_bit == 24 || audio->samp_bit == 32) {
            if (audio->samp_bit == 24) {
                factor = 8388608.0;
            } else {
                factor = 2147483648.0;
            }
            for (k = nread - 1; k >= 0; k--) {
                data[k] = weight * (double)ldata[k] / factor;
            }
        } else if (audio->samp_bit == 16 || audio->samp_bit == 8) {
            for (k = nread - 1; k >= 0; k--) {
                data[k] = weight * (double)sdata[k] / 32768.0;
            }
        } else if (audio->samp_bit == 33) {
            for (k = nread - 1; k >= 0; k--) {
                data[k] = weight * (double)fdata[k];
            }
        } else {
            if (weight != 1.0) {
                for (k = nread - 1; k >= 0; k--) {
                    data[k] *= weight;
                }
            }
        }
    }
    
    return nread;
}

long spReadAudioDouble(spAudio audio, double *data, long length)
{
    return spReadAudioDoubleWeighted(audio, data, length, 1.0);
}

long spWriteAudioDoubleWeighted(spAudio audio, double *data, long length, double weight)
{
    long nwrite = 0;
    long offset;
    double factor;

    if (audio->internal_buffer == NULL) return -1;
    
    factor = spGetBitNormalizeFactor(audio->samp_bit);
    
    spDebug(80, "spWriteAudioDoubleWeighted", "length = %ld, weight = %f, factor = %f\n",
            length, weight, factor);

    offset = 0;
    while (offset < length) {
        nwrite = MIN(audio->internal_buffer_size / spGetBitByte(audio->samp_bit, NULL), length - offset);
        spDebug(100, "spWriteAudioDoubleWeighted", "offset = %ld, nwrite = %ld\n", offset, nwrite);
        spConvertDoubleToBitWeighted(data + offset, nwrite, audio->internal_buffer,
                                     audio->samp_bit, factor * weight);
        
        if ((nwrite = audio->func_list.write(audio, audio->internal_buffer, nwrite)) <= 0) {
            spDebug(10, "spWriteAudioDoubleWeighted", "write failed: nwrite = %ld\n", nwrite);
            break;
        }
        offset += nwrite;
    }
    
    spDebug(80, "spWriteAudioDoubleWeighted", "done: offset = %ld, nwrite = %ld\n", offset, nwrite);
    
    return offset;
}

long spWriteAudioDouble(spAudio audio, double *data, long length)
{
    return spWriteAudioDoubleWeighted(audio, data, length, 1.0);
}

spBool spGetAudioOutputPosition(spAudio audio, spLong *position)
{
    if (audio == NULL) return SP_FALSE;

    return audio->func_list.get_output_position(audio, position);
}

spBool spStopAudio(spAudio audio)
{
    if (audio == NULL) return SP_FALSE;

    audio->input_total = 0;
    audio->output_total = 0;
    
    return audio->func_list.stop(audio);
}

spBool spSyncAudio(spAudio audio)
{
    if (audio == NULL) return SP_FALSE;

    audio->input_total = 0;
    audio->output_total = 0;

    return audio->func_list.sync(audio);
}

void _spFreeAudioDriver(spAudio audio)
{
    spDebug(20, "spFreeAudioDriver", "in\n");

    if (audio != NULL) {
        audio->func_list.free_driver(audio);
        spDebug(20, "spFreeAudioDriver", "free_driver done\n");

        spRemoveExitCallback((spExitCallbackFunc)spTerminateAudio, audio);
        spDebug(20, "spFreeAudioDriver", "spRemoveExitCallback done\n");
        
        if (audio->internal_buffer != NULL) {
            xfree(audio->internal_buffer);
        }
        xfree(audio);
    }
    
    spDebug(20, "spFreeAudioDriver", "done\n");
    
    return;
}

void spTerminateAudio(spAudio audio)
{
    if (audio != NULL) {
        spDebug(20, "spTerminateAudio", "audio != NULL\n");

        spCloseAudioDevice(audio);
        audio->func_list.terminate();
        spFreeAudioDriver(audio);
    }

    spDebug(20, "spTerminateAudio", "done\n");

    return;
}

spBool spAddAudioExitCallback(spAudio audio, spExitCallbackFunc callback)
{
    spRemoveExitCallback((spExitCallbackFunc)spTerminateAudio, audio);
    
    return spAddExitCallback(callback, audio);
}
