/*
 *        spAudio_Android.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spFile.h>
#include <sp/spThread.h>

#include <sp/spAudioP.h>

#if defined(SP_SUPPORT_AUDIO) && defined(ANDROID)
#include <SLES/OpenSLES.h>
#include <SLES/OpenSLES_Android.h>

// this callback handler is called every time a buffer finishes playing
static void bqPlayerCallback(SLAndroidSimpleBufferQueueItf bq, void *context)
{
    spAudio audio;
    SLresult result;

    spDebug(100, "bqPlayerCallback", "in\n");
    
    audio = (spAudio)context;
    
    if (audio == NULL || bq != audio->bqPlayerBufferQueue || audio->doStop == SP_TRUE) return;
    
    spDebug(100, "bqPlayerCallback", "before lock: playerFillCount = %ld\n", audio->playerFillCount);
    spLockMutex(audio->playerMutex);
    audio->playerFillCount = MAX(audio->playerFillCount - 1, 0);
    if (audio->playerFillCount < audio->playerNumBlock) {
        spDebug(100, "bqPlayerCallback", "call spSetEvent\n");
        spSetEvent(audio->playerEvent);
    }
    spDebug(100, "bqPlayerCallback", "playerBufferOffset = %ld, playerCopyOffset = %ld\n",
            audio->playerBufferOffset, audio->playerCopyOffset);
    audio->playerBufferOffset += audio->playerBlockSize;
    audio->playerTotalWrite += audio->playerBlockSize;
    if (audio->playerBufferOffset >= audio->playerBufferSize) {
        audio->playerBufferOffset = 0;
    }
    spUnlockMutex(audio->playerMutex);

    if (audio->call_func != NULL && (audio->call_type & SP_AUDIO_OUTPUT_POSITION_CALLBACK)) {
        spLong position;
        
        if (spGetAudioOutputPosition(audio, &position) == SP_TRUE) {
            if (audio->call_func(audio, SP_AUDIO_OUTPUT_POSITION_CALLBACK,
                                 &position, NULL, audio->call_data) == SP_FALSE) {
                audio->doStop = SP_TRUE;
            }
        }
    }

    spDebug(100, "bqPlayerCallback", "done\n");
}

static spBool spAudioRecorderEnqueue(spAudio audio, long offset)
{
    SLresult result;
    
    // enqueue buffer
    result = (*audio->recorderBufferQueue)->Enqueue(audio->recorderBufferQueue,
                                                    audio->recorderBuffer + offset,
                                                    audio->recorderBlockSize);
    if (SL_RESULT_SUCCESS != result) {
        spDebug(1, "spAudioRecorderEnqueue", "Enqueue failed: result = %d\n", result);
        return SP_FALSE;
    }

    spDebug(80, "spAudioRecorderEnqueue", "done\n");
    
    return SP_TRUE;
}

// this callback handler is called every time a buffer finishes recording
static void bqRecorderCallback(SLAndroidSimpleBufferQueueItf bq, void *context)
{
    spAudio audio;
    SLresult result;
    
    audio = (spAudio)context;
    
    if (audio == NULL || bq != audio->recorderBufferQueue || audio->doStop == SP_TRUE) return;
    
    spLockMutex(audio->recorderMutex);
    spDebug(100, "bqRecorderCallback", "recorderFillCount = %ld, recorderNumBlock = %ld\n",
            audio->recorderFillCount, audio->recorderNumBlock);
    
    if (audio->recorderFillCount < audio->recorderNumBlock) {
        audio->recorderFillCount = audio->recorderFillCount + 1;
        spDebug(100, "bqRecorderCallback", "recorderBufferOffset = %ld, recorderCopyOffset = %ld\n",
                audio->recorderBufferOffset, audio->recorderCopyOffset);
        audio->recorderBufferOffset += audio->recorderBlockSize;
        if (audio->recorderBufferOffset >= audio->recorderBufferSize) {
            audio->recorderBufferOffset = 0;
        }
    }
    spAudioRecorderEnqueue(audio, audio->recorderBufferOffset);
    spSetEvent(audio->recorderEvent);
    spUnlockMutex(audio->recorderMutex);
    
    spDebug(100, "bqRecorderCallback", "done\n");
}

int spGetNumAudioDriverArch(void)
{
    return 1;
}

const char *spGetAudioDriverNameArch(int index)
{
    if (index == 0) {
        return "Default Device";
    }
    
    return NULL;
}

spBool spInitAudioDriverArch(spAudio audio, const char *driver_name, void *data)
{
    SLresult result;
    const SLEngineOption engineOption[] = {SL_ENGINEOPTION_THREADSAFE, SL_BOOLEAN_TRUE};
    const SLInterfaceID ids[1] = {SL_IID_VOLUME};
    const SLboolean req[1] = {SL_BOOLEAN_FALSE};

    spDebug(80, "spInitAudioDriverArch", "in\n");
    
    // create engine
    result = slCreateEngine(&audio->engineObject, 1, engineOption, 0, NULL, NULL);
    if (SL_RESULT_SUCCESS != result) {
        spDebug(1, "spInitAudioDriverArch", "slCreateEngine failed: result = %d\n", result);
        return SP_FALSE;
    }

    // realize the engine
    result = (*audio->engineObject)->Realize(audio->engineObject, SL_BOOLEAN_FALSE);
    if (SL_RESULT_SUCCESS != result) {
        spDebug(1, "spInitAudioDriverArch", "Realize engineObject failed: result = %d\n", result);
        return SP_FALSE;
    }

    // get the engine interface, which is needed in order to create other objects
    result = (*audio->engineObject)->GetInterface(audio->engineObject, SL_IID_ENGINE, &audio->engineEngine);
    if (SL_RESULT_SUCCESS != result) {
        spDebug(1, "spInitAudioDriverArch", "GetInterface of SL_IID_ENGINE failed: result = %d\n", result);
        return SP_FALSE;
    }
    
    // create output mix
    result = (*audio->engineEngine)->CreateOutputMix(audio->engineEngine, &audio->outputMixObject, 1, ids, req);
    if (SL_RESULT_SUCCESS != result) {
        spDebug(1, "spInitAudioDriverArch", "CreateOutputMix failed: result = %d\n", result);
        return SP_FALSE;
    }
    result = (*audio->outputMixObject)->Realize(audio->outputMixObject, SL_BOOLEAN_FALSE);
    if (SL_RESULT_SUCCESS != result) {
        spDebug(1, "spInitAudioDriverArch", "Realize outputMixObject failed: result = %d\n", result);
        return SP_FALSE;
    }

    audio->playerMutex = spCreateMutex(NULL);
    audio->playerEvent = spCreateEvent(SP_TRUE, SP_TRUE);
    audio->recorderMutex = spCreateMutex(NULL);
    audio->recorderEvent = spCreateEvent(SP_TRUE, SP_TRUE);

    spDebug(80, "spInitAudioDriverArch", "done\n");
    
    return SP_TRUE;
}

spBool spGetNumAudioDeviceArch(spAudio audio, int *num_device)
{
    *num_device = 1;
    return SP_TRUE;
}

const char *spGetAudioDeviceNameArch(spAudio audio, int device_index)
{
    if (device_index == 0) {
        return "Default";
    }
    return NULL;
}

spBool spSelectAudioDeviceArch(spAudio audio, int device_index)
{
    if (device_index == 0) {
        return SP_TRUE;
    }
    return SP_FALSE;
}

spBool spSetAudioSampleBitArch(spAudio audio)
{
    audio->samp_bit = 16;
    return SP_TRUE;
}

spBool spSetAudioSampleRateArch(spAudio audio)
{
    return SP_TRUE;
}

spBool spSetAudioChannelArch(spAudio audio)
{
    return SP_TRUE;
}

spBool spSetAudioBufferSizeArch(spAudio audio)
{
    return SP_TRUE;
}

spBool spSetAudioNumBufferArch(spAudio audio)
{
    audio->num_buffer = MIN(audio->num_buffer, 128);
    return SP_TRUE;
}

spBool spSetAudioBlockingModeArch(spAudio audio)
{
    return SP_TRUE;
}

spAudioCallbackType spSetAudioCallbackFuncArch(spAudio audio)
{
    return (audio->call_type & SP_AUDIO_OUTPUT_POSITION_CALLBACK);
}

static void spAudioRecorderEnqueueAll(spAudio audio)
{
    audio->recorderBufferOffset = 0;
    
    while (1) {
        spAudioRecorderEnqueue(audio, audio->recorderBufferOffset);
        audio->recorderBufferOffset += audio->recorderBlockSize;
        if (audio->recorderBufferOffset >= audio->recorderBufferSize) {
            audio->recorderBufferOffset = 0;
            break;
        }
    }

    return;
}

spBool spOpenInputAudioDeviceArch(spAudio audio)
{
    SLresult result;
    // configure audio source
    SLDataLocator_IODevice loc_dev = {SL_DATALOCATOR_IODEVICE, SL_IODEVICE_AUDIOINPUT,
            SL_DEFAULTDEVICEID_AUDIOINPUT, NULL};
    SLDataSource audioSrc = {&loc_dev, NULL};

    // configure audio sink
    SLDataLocator_AndroidSimpleBufferQueue loc_bq = {SL_DATALOCATOR_ANDROIDSIMPLEBUFFERQUEUE, 2};
    SLDataFormat_PCM format_pcm = {SL_DATAFORMAT_PCM, 1, SL_SAMPLINGRATE_8,
        SL_PCMSAMPLEFORMAT_FIXED_16, SL_PCMSAMPLEFORMAT_FIXED_16,
        SL_SPEAKER_FRONT_CENTER, SL_BYTEORDER_LITTLEENDIAN};
    SLDataSink audioSnk = {&loc_bq, &format_pcm};
    // create audio recorder
    // (requires the RECORD_AUDIO permission)
    const SLInterfaceID ids[1] = {SL_IID_ANDROIDSIMPLEBUFFERQUEUE};
    const SLboolean req[1] = {SL_BOOLEAN_TRUE};

    spDebug(80, "spOpenInputAudioDeviceArch", "in\n");
    
    audio->recorderNumBlock = MAX(audio->num_buffer, 2);
    loc_bq.numBuffers = audio->recorderNumBlock;
    
    if (audio->num_channel == 2) {
        format_pcm.channelMask = SL_SPEAKER_FRONT_LEFT | SL_SPEAKER_FRONT_RIGHT;
        format_pcm.numChannels = 2;
    }
    format_pcm.samplesPerSec = (SLuint32)spRound(audio->samp_rate * 1000.0);
    
    result = (*audio->engineEngine)->CreateAudioRecorder(audio->engineEngine, &audio->recorderObject, &audioSrc,
                                                         &audioSnk, 1, ids, req);
    if (SL_RESULT_SUCCESS != result) {
        return SP_FALSE;
    }

    // realize the audio recorder
    result = (*audio->recorderObject)->Realize(audio->recorderObject, SL_BOOLEAN_FALSE);
    if (SL_RESULT_SUCCESS != result) {
        return SP_FALSE;
    }

    // get the record interface
    result = (*audio->recorderObject)->GetInterface(audio->recorderObject, SL_IID_RECORD, &audio->recorderRecord);
    if (SL_RESULT_SUCCESS != result) {
        return SP_FALSE;
    }

    // get the buffer queue interface
    result = (*audio->recorderObject)->GetInterface(audio->recorderObject, SL_IID_ANDROIDSIMPLEBUFFERQUEUE,
                                                    &audio->recorderBufferQueue);
    if (SL_RESULT_SUCCESS != result) {
        return SP_FALSE;
    }

    // register callback on the buffer queue
    result = (*audio->recorderBufferQueue)->RegisterCallback(audio->recorderBufferQueue, bqRecorderCallback,
                                                             audio);
    if (SL_RESULT_SUCCESS != result) {
        return SP_FALSE;
    }

    audio->recorderBlockSize = audio->buffer_size;
    audio->recorderBufferSize = audio->buffer_size * audio->recorderNumBlock;
    spDebug(80, "spOpenInputAudioDeviceArch", "audio->recorderBlockSize = %ld, audio->recorderBufferSize = %ld\n",
            audio->recorderBlockSize, audio->recorderBufferSize);
    
    if (audio->recorderBuffer != NULL) xfree(audio->recorderBuffer);
    audio->recorderBuffer = xalloc(audio->recorderBufferSize, char);
    if (audio->recorderBuffer == NULL) {
        spDebug(1, "spOpenInputAudioDeviceArch", "Cannot allocate buffer memory\n");
        return SP_FALSE;
    }
    
    memset(audio->recorderBuffer, 0, audio->recorderBufferSize);
    audio->recorderStarted = SP_FALSE;
    audio->recorderFillCount = 0;
    audio->recorderBufferOffset = 0;
    audio->recorderCopyOffset = 0;
    audio->doStop = SP_FALSE;
    (*audio->recorderRecord)->SetRecordState(audio->recorderRecord, SL_RECORDSTATE_PAUSED);
    (*audio->recorderBufferQueue)->Clear(audio->recorderBufferQueue);

    spAudioRecorderEnqueueAll(audio);
    spResetEvent(audio->recorderEvent);

    spDebug(80, "spOpenInputAudioDeviceArch", "done\n");
    
    return SP_TRUE;
}

spBool spOpenOutputAudioDeviceArch(spAudio audio)
{
    SLresult result;
    // configure audio source
    SLDataLocator_AndroidSimpleBufferQueue loc_bufq = {SL_DATALOCATOR_ANDROIDSIMPLEBUFFERQUEUE, 2};
    SLDataFormat_PCM format_pcm = {SL_DATAFORMAT_PCM, 1, SL_SAMPLINGRATE_8,
        SL_PCMSAMPLEFORMAT_FIXED_16, SL_PCMSAMPLEFORMAT_FIXED_16,
        SL_SPEAKER_FRONT_CENTER, SL_BYTEORDER_LITTLEENDIAN};
    SLDataSource audioSrc = {&loc_bufq, &format_pcm};
    SLDataLocator_OutputMix loc_outmix = {SL_DATALOCATOR_OUTPUTMIX, NULL};
    SLDataSink audioSnk = {&loc_outmix, NULL};
    const SLInterfaceID ids[1] = {SL_IID_BUFFERQUEUE};
    const SLboolean req[1] = {SL_BOOLEAN_TRUE};

    spDebug(80, "spOpenOutputAudioDeviceArch", "in\n");
    
    audio->playerNumBlock = MAX(audio->num_buffer, 2);
    loc_bufq.numBuffers = audio->playerNumBlock;
    
    if (audio->num_channel == 2) {
        format_pcm.channelMask = SL_SPEAKER_FRONT_LEFT | SL_SPEAKER_FRONT_RIGHT;
        format_pcm.numChannels = 2;
    }
    format_pcm.samplesPerSec = (SLuint32)spRound(audio->samp_rate * 1000.0);

    loc_outmix.outputMix = audio->outputMixObject;
    
    result = (*audio->engineEngine)->CreateAudioPlayer(audio->engineEngine, &audio->bqPlayerObject,
                                                       &audioSrc, &audioSnk, 1, ids, req);
    if (SL_RESULT_SUCCESS != result) {
        spDebug(1, "spOpenOutputAudioDeviceArch", "CreateAudioPlayer failed: result = %d\n", result);
        return SP_FALSE;
    }

    // realize the player
    result = (*audio->bqPlayerObject)->Realize(audio->bqPlayerObject, SL_BOOLEAN_FALSE);
    if (SL_RESULT_SUCCESS != result) {
        spDebug(1, "spOpenOutputAudioDeviceArch", "Realize failed: result = %d\n", result);
        return SP_FALSE;
    }

    // get the play interface
    result = (*audio->bqPlayerObject)->GetInterface(audio->bqPlayerObject, SL_IID_PLAY, &audio->bqPlayerPlay);
    if (SL_RESULT_SUCCESS != result) {
        spDebug(1, "spOpenOutputAudioDeviceArch", "GetInterface of SL_IID_PLAY failed: result = %d\n", result);
        return SP_FALSE;
    }

    // get the buffer queue interface
    result = (*audio->bqPlayerObject)->GetInterface(audio->bqPlayerObject, SL_IID_BUFFERQUEUE,
                                                    &audio->bqPlayerBufferQueue);
    if (SL_RESULT_SUCCESS != result) {
        spDebug(1, "spOpenOutputAudioDeviceArch", "GetInterface of SL_IID_BUFFERQUEUE failed: result = %d\n", result);
        return SP_FALSE;
    }

    // register callback on the buffer queue
    result = (*audio->bqPlayerBufferQueue)->RegisterCallback(audio->bqPlayerBufferQueue,
                                                             bqPlayerCallback, audio);
    if (SL_RESULT_SUCCESS != result) {
        spDebug(1, "spOpenOutputAudioDeviceArch", "RegisterCallback failed: result = %d\n", result);
        return SP_FALSE;
    }

    audio->playerBlockSize = audio->buffer_size;
    audio->playerBufferSize = audio->buffer_size * audio->playerNumBlock;
    if (audio->playerBuffer != NULL) xfree(audio->playerBuffer);
    audio->playerBuffer = xalloc(audio->playerBufferSize, char);
    if (audio->playerBuffer == NULL) {
        spDebug(1, "spOpenOutputAudioDeviceArch", "Cannot allocate buffer memory\n");
        return SP_FALSE;
    }
    
    memset(audio->playerBuffer, 0, audio->playerBufferSize);
    audio->playerStarted = SP_FALSE;
    audio->playerFillCount = 0;
    audio->playerBufferOffset = 0;
    audio->playerCopyOffset = 0;
    audio->playerTotalWrite = 0;
    audio->doStop = SP_FALSE;
    spSetEvent(audio->playerEvent);

    spDebug(80, "spOpenOutputAudioDeviceArch", "done\n");
    
    return SP_TRUE;
}

spBool spCloseAudioDeviceArch(spAudio audio)
{
    spDebug(80, "spCloseAudioDeviceArch", "in\n");
    
    if (audio->bqPlayerObject != NULL) {
        if (audio->playerStarted == SP_TRUE) {
            while (audio->playerFillCount > 0) {
                spWaitEvent(audio->playerEvent);
                spLockMutex(audio->playerMutex);
                spResetEvent(audio->playerEvent);
                spUnlockMutex(audio->playerMutex);
            }
        }
        audio->playerStarted = SP_FALSE;
        
        if (audio->bqPlayerPlay != NULL) {
            (*audio->bqPlayerPlay)->SetPlayState(audio->bqPlayerPlay, SL_PLAYSTATE_STOPPED);
        }
        (*audio->bqPlayerObject)->Destroy(audio->bqPlayerObject);
        audio->bqPlayerObject = NULL;
        audio->bqPlayerPlay = NULL;
        audio->bqPlayerBufferQueue = NULL;
    }
    
    if (audio->playerBuffer != NULL) {
        xfree(audio->playerBuffer);
        audio->playerBuffer = NULL;
    }
    
    spDebug(80, "spCloseAudioDeviceArch", "before spLockMutex of recorderMutex\n");
    spLockMutex(audio->recorderMutex);
    audio->recorderStarted = SP_FALSE;
    if (audio->recorderObject != NULL) {
        if (audio->recorderRecord != NULL) {
            (*audio->recorderRecord)->SetRecordState(audio->recorderRecord, SL_RECORDSTATE_STOPPED);
        }
        (*audio->recorderObject)->Destroy(audio->recorderObject);
        audio->recorderObject = NULL;
        audio->recorderRecord = NULL;
        audio->recorderBufferQueue = NULL;
    }
    if (audio->recorderBuffer != NULL) {
        xfree(audio->recorderBuffer);
        audio->recorderBuffer = NULL;
    }
    spUnlockMutex(audio->recorderMutex);
    
    spDebug(80, "spCloseAudioDeviceArch", "done\n");
    
    return SP_TRUE;
}

long spReadAudioArch(spAudio audio, char *data, long length)
{
    SLresult result;
    long dataOffset;
    long blockOffset;
    long remainSize;
    long blockSize;

    if (audio->recorderStarted == SP_FALSE) {
        audio->recorderStarted = SP_TRUE;
        result = (*audio->recorderRecord)->SetRecordState(audio->recorderRecord, SL_RECORDSTATE_RECORDING);
        if (SL_RESULT_SUCCESS != result) {
            spDebug(1, "spReadAudioArch", "SetRecordState failed: result = %d\n", result);
            return -1;
        }
        spDebug(80, "spReadAudioArch", "SetRecordState done\n");
    }

    dataOffset = 0;
    remainSize = length * audio->samp_byte;

    while (remainSize > 0 && audio->doStop == SP_FALSE) {
        spDebug(100, "spReadAudioArch", "remainSize = %ld, recorderCopyOffset = %ld, recorderBufferOffset = %ld, recorderFillCount = %ld\n",
                remainSize, audio->recorderCopyOffset, audio->recorderBufferOffset, audio->recorderFillCount);

        spWaitEvent(audio->recorderEvent);
        spDebug(100, "spReadAudioArch", "wait event done\n");
        
        if (dataOffset == 0) {
            blockOffset = audio->recorderCopyOffset % audio->recorderBlockSize;
        } else {
            blockOffset = 0;
        }
        blockSize = MIN(audio->recorderBlockSize - blockOffset, remainSize);

        spDebug(100, "spReadAudioArch", "before spLockMutex of recorderMutex\n");
        spLockMutex(audio->recorderMutex);
        memcpy(data + dataOffset, audio->recorderBuffer + audio->recorderCopyOffset, blockSize);

        spDebug(100, "spReadAudioArch", "blockOffset = %ld, blockSize = %ld, recorderBlockSize = %ld\n",
                blockOffset, blockSize, audio->recorderBlockSize);
        
        if (blockOffset + blockSize == audio->recorderBlockSize) {
            spDebug(100, "spReadAudioArch", "recorderFillCount = %ld\n", audio->recorderFillCount);
            
            --audio->recorderFillCount;
            if (audio->recorderFillCount <= 0) {
                spResetEvent(audio->recorderEvent);
                audio->recorderFillCount = 0;
            }
        }
        audio->recorderCopyOffset += blockSize;
        if (audio->recorderCopyOffset >= audio->recorderBufferSize) {
            audio->recorderCopyOffset = 0;
        }
        spUnlockMutex(audio->recorderMutex);
        
        dataOffset += blockSize;
        remainSize -= blockSize;
    }
    
    spDebug(80, "spReadAudioArch", "done: dataOffset = %ld, length = %ld\n", dataOffset, length);
    
    return dataOffset / audio->samp_byte;
}

static spBool spAudioPlayerEnqueue(spAudio audio, long offset)
{
    SLresult result;
    
    // enqueue buffer
    spDebug(80, "spAudioPlayerEnqueue", "before Enqueue lock\n");
    spLockMutex(audio->playerMutex);
    audio->playerFillCount = audio->playerFillCount + 1;
    if (audio->playerFillCount >= audio->playerNumBlock) {
        spResetEvent(audio->playerEvent);
    }
    result = (*audio->bqPlayerBufferQueue)->Enqueue(audio->bqPlayerBufferQueue,
                                                    audio->playerBuffer + offset,
                                                    audio->playerBlockSize);
    spUnlockMutex(audio->playerMutex);
    if (SL_RESULT_SUCCESS != result) {
        spDebug(1, "spAudioPlayerEnqueue", "Enqueue failed: result = %d\n", result);
        return SP_FALSE;
    }
    
    if (audio->playerStarted == SP_FALSE) {
        spDebug(80, "spAudioPlayerEnqueue", "before SetPlayState lock\n");
        spLockMutex(audio->playerMutex);
        if (audio->playerFillCount >= MIN(/*1*/2, audio->playerNumBlock)) {
            spDebug(50, "spAudioPlayerEnqueue", "SetPlayState to SL_PLAYSTATE_PLAYING\n");
            spResetEvent(audio->playerEvent);
            result = (*audio->bqPlayerPlay)->SetPlayState(audio->bqPlayerPlay, SL_PLAYSTATE_PLAYING);
            if (SL_RESULT_SUCCESS != result) {
                spUnlockMutex(audio->playerMutex);
                spDebug(1, "spAudioPlayerEnqueue", "SetPlayState failed: result = %d\n", result);
                return SP_FALSE;
            }
            audio->playerStarted = SP_TRUE;
        }
        spUnlockMutex(audio->playerMutex);
    }

    spDebug(80, "spAudioPlayerEnqueue", "done\n");
    
    return SP_TRUE;
}

long spWriteAudioArch(spAudio audio, char *data, long length)
{
    SLresult result;
    long dataOffset;
    long blockOffset;
    long remainSize;
    long blockSize;

    dataOffset = 0;
    remainSize = length * audio->samp_byte;

    while (remainSize > 0) {
        spDebug(100, "spWriteAudioArch", "remainSize = %ld, playerCopyOffset = %ld, playerBufferOffset = %ld, playerTotalWrite = %ld\n",
                remainSize, audio->playerCopyOffset, audio->playerBufferOffset, (long)audio->playerTotalWrite);
        spWaitEvent(audio->playerEvent);
        spDebug(100, "spWriteAudioArch", "wait event done\n");
        
        if (dataOffset == 0) {
            blockOffset = audio->playerCopyOffset % audio->playerBlockSize;
        } else {
            blockOffset = 0;
        }
        blockSize = MIN(audio->playerBlockSize - blockOffset, remainSize);

        memcpy(audio->playerBuffer + audio->playerCopyOffset, data + dataOffset, blockSize);
        spDebug(100, "spWriteAudioArch", "blockOffset = %ld, blockSize = %ld, playerBlockSize = %ld\n",
                blockOffset, blockSize, audio->playerBlockSize);
        if (blockOffset + blockSize == audio->playerBlockSize) {
            if (spAudioPlayerEnqueue(audio, audio->playerCopyOffset - blockOffset) == SP_FALSE) {
                return -1;
            }
        }

        dataOffset += blockSize;
        remainSize -= blockSize;
        
        spDebug(80, "spWriteAudioArch", "update playerCopyOffset from %ld, blockSize = %ld\n", audio->playerCopyOffset, blockSize);
        spLockMutex(audio->playerMutex);
        audio->playerCopyOffset += blockSize;
        if (audio->playerCopyOffset >= audio->playerBufferSize) {
            audio->playerCopyOffset = 0;
        }
        spUnlockMutex(audio->playerMutex);
    }
    
    spDebug(80, "spWriteAudioArch", "done: dataOffset = %ld, length = %ld\n", dataOffset, length);
    
    return dataOffset / audio->samp_byte;
}

spBool spGetAudioOutputPositionArch(spAudio audio, spLong *position)
{
    spDebug(100, "spGetAudioOutputPositionArch", "playerTotalWrite = %ld\n", (long)audio->playerTotalWrite);
    
    *position = audio->playerTotalWrite / (spLong)(audio->samp_byte * audio->num_channel);
    
    return SP_TRUE;
}

spBool spStopAudioArch(spAudio audio)
{
    audio->doStop = SP_TRUE;
    if (audio->bqPlayerObject != NULL) {
        if (audio->bqPlayerPlay != NULL) {
            (*audio->bqPlayerPlay)->SetPlayState(audio->bqPlayerPlay, SL_PLAYSTATE_STOPPED);
        }
    }
    if (audio->recorderObject != NULL) {
        if (audio->recorderRecord != NULL) {
            (*audio->recorderRecord)->SetRecordState(audio->recorderRecord, SL_RECORDSTATE_STOPPED);
        }
    }
    audio->playerTotalWrite = 0;
    
    return SP_TRUE;
}

spBool spSyncAudioArch(spAudio audio)
{
    SLresult result;
    
    spDebug(50, "spSyncAudioArch", "in\n");
    
    if (audio->bqPlayerObject != NULL) {
        audio->playerStarted = SP_FALSE;
        (*audio->bqPlayerPlay)->SetPlayState(audio->bqPlayerPlay, SL_PLAYSTATE_PAUSED);
        (*audio->bqPlayerBufferQueue)->Clear(audio->bqPlayerBufferQueue);
    
        spLockMutex(audio->playerMutex);
        memset(audio->playerBuffer, 0, audio->playerBufferSize);
        audio->playerCopyOffset = 0;
        audio->playerBufferOffset = 0;
        audio->playerTotalWrite = 0;
        spSetEvent(audio->playerEvent);
        spUnlockMutex(audio->playerMutex);
    }
    
    if (audio->recorderObject != NULL) {
        if (audio->recorderStarted == SP_TRUE) {
            audio->recorderStarted = SP_FALSE;
            (*audio->recorderRecord)->SetRecordState(audio->recorderRecord, SL_RECORDSTATE_PAUSED);
            (*audio->recorderBufferQueue)->Clear(audio->recorderBufferQueue);
        
            spLockMutex(audio->recorderMutex);
            memset(audio->recorderBuffer, 0, audio->recorderBufferSize);
            audio->recorderCopyOffset = 0;
            audio->recorderBufferOffset = 0;
            spResetEvent(audio->recorderEvent);
            spUnlockMutex(audio->recorderMutex);
        
            spAudioRecorderEnqueueAll(audio);
        }
    }
    
    spDebug(50, "spSyncAudioArch", "done\n");
    
    return SP_TRUE;
}

spBool spFreeAudioDriverArch(spAudio audio)
{
    if (audio->playerMutex != NULL) {
        spDestroyMutex(audio->playerMutex);
        audio->playerMutex = NULL;
    }
    if (audio->playerEvent != NULL) {
        spDestroyEvent(audio->playerEvent);
        audio->playerEvent = NULL;
    }
    if (audio->recorderMutex != NULL) {
        spDestroyMutex(audio->recorderMutex);
        audio->recorderMutex = NULL;
    }
    if (audio->recorderEvent != NULL) {
        spDestroyEvent(audio->recorderEvent);
        audio->recorderEvent = NULL;
    }
    
    if (audio->outputMixObject != NULL) {
        (*audio->outputMixObject)->Destroy(audio->outputMixObject);
        audio->outputMixObject = NULL;
    }
    if (audio->engineObject != NULL) {
        (*audio->engineObject)->Destroy(audio->engineObject);
        audio->engineObject = NULL;
    }
    
    return SP_TRUE;
}

void spTerminateAudioArch(void)
{
    return;
}
#endif
