/*
 *        spAudio_Win.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spFile.h>

#include <sp/spWaveP.h>
#include <sp/spAudioP.h>

#if defined(SP_SUPPORT_AUDIO) && defined(_WIN32)
#include <stdio.h>
#include <stdlib.h>

#include "spAudio_WinArch.h"

typedef struct _spWinDeviceList
{
    UINT output_device_id;
    UINT input_device_id;
    
    WAVEOUTCAPS woc;
    WAVEINCAPS wic;
    
    struct _spWinDeviceList *prev;
    struct _spWinDeviceList *next;
} *spWinDeviceList;

static spWinDeviceList initDeviceList(void)
{
    spWinDeviceList list;
    
    list = xalloc(1, struct _spWinDeviceList);
    memset(list, 0, sizeof(struct _spWinDeviceList));
    list->output_device_id = 0;
    list->input_device_id = 0;

    return list;
}

static spWinDeviceList getDeviceList(int *pnum_device)
{
    static int num_device = 0;
    static spWinDeviceList first_list = NULL;

    if (first_list == NULL) {
        UINT device_id;
        UINT k, l;
        UINT num_input_device = 0;
        UINT num_output_device = 0;
        WAVEINCAPS wic;
        WAVEOUTCAPS woc;
        spWinDeviceList list, prev_list;
        spWinDeviceList olist;
        spBool same_found;

        prev_list = NULL;

        num_device = 0;
        num_output_device = waveInGetNumDevs() + 1;
        spDebug(10, "getDeviceList", "num_output_device = %d\n", num_output_device);
        for (k = 0; k < num_output_device; k++) {
            if (k == 0) {
                device_id = WAVE_MAPPER;
            } else {
                device_id = k - 1;
            }
            
            if (waveOutGetDevCaps(device_id, &woc, sizeof(WAVEOUTCAPS)) == MMSYSERR_NOERROR) {
                spDebug(10, "getDeviceList", "%d: woc.szPname = %s\n", device_id, woc.szPname);

                list = initDeviceList();
                list->output_device_id = device_id;
                memcpy(&list->woc, &woc, sizeof(WAVEOUTCAPS));

                if (prev_list == NULL) {
                    first_list = list;
                } else {
                    list->prev = prev_list;
                    prev_list->next = list;
                }
                prev_list = list;
                num_device++;
            }
        }

        num_output_device = (UINT)num_device;
        num_input_device = waveInGetNumDevs() + 1;
        spDebug(10, "getDeviceList", "num_output_device = %d, num_input_device = %d\n",
                num_output_device, num_input_device);
        
        for (k = 0; k < num_input_device; k++) {
            if (k == 0) {
                device_id = WAVE_MAPPER;
            } else {
                device_id = k - 1;
            }
            
            if (waveInGetDevCaps(device_id, &wic, sizeof(WAVEINCAPS)) == MMSYSERR_NOERROR) {
                spDebug(10, "getDeviceList", "%d: wic.szPname = %s\n", device_id, wic.szPname);

                olist = first_list;
                same_found = SP_FALSE;
                for (l = 0; l < num_output_device; l++) {
                    if (olist == NULL) {
                        break;
                    }
                    if (streq(olist->woc.szPname, wic.szPname)) {
                        spDebug(10, "getDeviceList", "device %d is same as %d\n",
                                device_id, olist->output_device_id);
                        olist->input_device_id = device_id;
                        memcpy(&olist->wic, &wic, sizeof(WAVEINCAPS));

                        same_found = SP_TRUE;
                    }
                    olist = olist->next;
                }

                if (same_found == SP_FALSE) {
                    list = initDeviceList();
                    list->input_device_id = device_id;
                    memcpy(&list->wic, &wic, sizeof(WAVEINCAPS));
                
                    if (prev_list == NULL) {
                        first_list = list;
                    } else {
                        list->prev = prev_list;
                        prev_list->next = list;
                    }
                    prev_list = list;
                    num_device++;
                }
            }
        }
    }
    spDebug(10, "getDeviceList", "total found device = %ld\n", num_device);

    if (pnum_device != NULL) *pnum_device = num_device;
    
    return first_list;
}

int spGetNumAudioDriverArch(void)
{
    int num_device = 0;
    spWinDeviceList list;

    if ((list = getDeviceList(&num_device)) != NULL) {
        return num_device;
    } else {
        return 0;
    }
}

const char *spGetAudioDriverNameArch(int index)
{
    int k;
    spWinDeviceList list;
    char *name = NULL;

    list = getDeviceList(NULL);
    for (k = 0;; k++) {
        if (list == NULL) break;
        
        if (k == index) {
            if (!strnone(list->woc.szPname)) {
                name = list->woc.szPname;
            } else if (!strnone(list->wic.szPname)) {
                name = list->wic.szPname;
            }
            break;
        }
        list = list->next;
    }

    return name;
}

spBool spInitAudioDriverArch(spAudio audio, const char *driver_name, void *data)
{
    int k;
    spBool input_found, output_found;
    spWinDeviceList list;

    audio->input_device_id = WAVE_MAPPER;
    audio->output_device_id = WAVE_MAPPER;

    if (!strnone(driver_name)) {
        input_found = SP_FALSE;
        output_found = SP_FALSE;
        
        list = getDeviceList(NULL);
        for (k = 0;; k++) {
            if (list == NULL) break;
        
            if (input_found == SP_FALSE &&
                !strnone(list->wic.szPname) && streq(list->wic.szPname, driver_name)) {
                audio->input_device_id = list->input_device_id;
                input_found = SP_TRUE;
            }
            if (output_found == SP_FALSE &&
                !strnone(list->woc.szPname) && streq(list->woc.szPname, driver_name)) {
                audio->output_device_id = list->output_device_id;
                output_found = SP_TRUE;
            }
            list = list->next;
        }
    }
    
    audio->input_audio = NULL;
    audio->output_audio = NULL;
    audio->samp_byte = audio->samp_bit / 8;
    
    return SP_TRUE;
}

spBool spGetNumAudioDeviceArch(spAudio audio, int *num_device)
{
    *num_device = 1;
    return SP_TRUE;
}

const char *spGetAudioDeviceNameArch(spAudio audio, int device_index)
{
    if (device_index == 0) {
        return "Default";
    }
    return NULL;
}

spBool spSelectAudioDeviceArch(spAudio audio, int device_index)
{
    if (device_index == 0) {
        return SP_TRUE;
    }
    return SP_FALSE;
}

spBool spSetAudioSampleBitArch(spAudio audio)
{
    audio->samp_bit = MIN(audio->samp_bit, 16);
    
    return SP_TRUE;
}

spBool spSetAudioSampleRateArch(spAudio audio)
{
    return SP_TRUE;
}

spBool spSetAudioChannelArch(spAudio audio)
{
    return SP_TRUE;
}

spBool spSetAudioBufferSizeArch(spAudio audio)
{
    return SP_TRUE;
}

spBool spSetAudioNumBufferArch(spAudio audio)
{
#if 1
    audio->num_buffer = MAX(audio->num_buffer, /*32*/2);
#endif
    return SP_TRUE;
}

spBool spSetAudioBlockingModeArch(spAudio audio)
{
    return SP_TRUE;
}

static spBool audioCallback(spWinAudio waudio, spAudioCallbackType call_type,
                            void *data1, void *data2, void *user_data)
{
    spAudio audio = (spAudio)user_data;

    return audio->call_func(audio, call_type, data1, data2, audio->call_data);
}

spAudioCallbackType spSetAudioCallbackFuncArch(spAudio audio)
{
    return (audio->call_type &
            (SP_AUDIO_OUTPUT_POSITION_CALLBACK | SP_AUDIO_OUTPUT_BUFFER_CALLBACK));
}

spBool spOpenInputAudioDeviceArch(spAudio audio)
{
    spDebug(100, "spOpenInputAudioDeviceArch", "in\n");
    
    if (audio->input_audio != NULL) {
        spDebug(10, "spOpenInputAudioDeviceArch", "error: input_audio != NULL\n");
        return SP_FALSE;
    }

    audio->input_audio = spInitAudio_Win(audio->num_buffer, audio->buffer_size);
    spSetAudioCallbackFunc_Win(audio->input_audio, audio->call_type, audioCallback, audio);
    
    if (spOpenAudioDevice_Win(audio->input_audio, audio->input_device_id, audio->samp_bit, audio->num_channel,
                              audio->samp_rate, audio->block_mode == SP_AUDIO_NONBLOCKING,
                              SP_FALSE) == SP_FALSE) {
        spDebug(10, "spOpenInputAudioDeviceArch", "spOpenAudioDevice_Win failed\n");
        spFreeAudio_Win(audio->input_audio);
        audio->input_audio = NULL;
        return SP_FALSE;
    }

    spDebug(100, "spOpenInputAudioDeviceArch", "done\n");
    
    return SP_TRUE;
}

spBool spOpenOutputAudioDeviceArch(spAudio audio)
{
    spDebug(100, "spOpenOutputAudioDeviceArch", "in\n");
    
    if (audio->output_audio != NULL) {
        spDebug(10, "spOpenOutputAudioDeviceArch", "error: output_audio != NULL\n");
        return SP_FALSE;
    }

    audio->output_audio = spInitAudio_Win(audio->num_buffer, audio->buffer_size);
    spSetAudioCallbackFunc_Win(audio->output_audio, audio->call_type, audioCallback, audio);
    
    if (spOpenAudioDevice_Win(audio->output_audio, audio->output_device_id, audio->samp_bit, audio->num_channel,
                              audio->samp_rate, audio->block_mode == SP_AUDIO_NONBLOCKING,
                              SP_TRUE) == SP_FALSE) {
        spDebug(10, "spOpenOutputAudioDeviceArch", "spOpenAudioDevice_Win failed\n");
        spFreeAudio_Win(audio->output_audio);
        audio->output_audio = NULL;
        return SP_FALSE;
    }

    spDebug(100, "spOpenOutputAudioDeviceArch", "done\n");
    
    return SP_TRUE;
}

spBool spCloseAudioDeviceArch(spAudio audio)
{
    if (audio->input_audio == NULL && audio->output_audio == NULL) {
        return SP_FALSE;
    }
    
    if (audio->output_audio != NULL) {
        spCloseAudioDevice_Win(audio->output_audio);
        spFreeAudio_Win(audio->output_audio);
        audio->output_audio = NULL;
    }
    if (audio->input_audio != NULL) {
        spCloseAudioDevice_Win(audio->input_audio);
        spFreeAudio_Win(audio->input_audio);
        audio->input_audio = NULL;
    }

    return SP_TRUE;
}

long spReadAudioArch(spAudio audio, char *data, long length)
{
    if (audio->input_audio == NULL) return -1;
    
    return spReadWave_Win(audio->input_audio, data, length);
}

long spWriteAudioArch(spAudio audio, char *data, long length)
{
    if (audio->output_audio == NULL) return -1;
    
    return spWriteWave_Win(audio->output_audio, data, length);
}

spBool spGetAudioOutputPositionArch(spAudio audio, spLong *position)
{
    return spGetAudioOutputPosition_Win(audio->output_audio, position);
}

spBool spStopAudioArch(spAudio audio)
{
    spResetDevice_Win(audio->output_audio);
    spResetDevice_Win(audio->input_audio);
    
    return SP_TRUE;
}

spBool spSyncAudioArch(spAudio audio)
{
    spSyncDevice_Win(audio->output_audio);
    spSyncDevice_Win(audio->input_audio);

    return SP_TRUE;
}

spBool spFreeAudioDriverArch(spAudio audio)
{
    return SP_TRUE;
}

void spTerminateAudioArch(void)
{
    return;
}
#endif
